@@ -28,12 +28,12 @@ from zigpy_znp.types.nvids import NwkNvIds
 
 
 ZDO_ENDPOINT = 0
-ZDO_REQUEST_TIMEOUT = 15  # seconds
+ZDO_REQUEST_TIMEOUT = 10  # seconds
 DATA_CONFIRM_TIMEOUT = 5  # seconds
 LOGGER = logging.getLogger(__name__)
 
 
-ZDO_CONVERTERS = {
+ZDO_COMPLEX_CONVERTERS = {
     ZDOCmd.Node_Desc_req: (
         (
             lambda addr, device, NWKAddrOfInterest: c.ZDO.NodeDescReq.Req(
@@ -41,7 +41,16 @@ ZDO_CONVERTERS = {
             )
         ),
         (lambda addr: c.ZDO.NodeDescRsp.Callback(partial=True, Src=addr)),
-        (lambda rsp: (ZDOCmd.Node_Desc_rsp, [rsp.NodeDescriptor])),
+        (
+            lambda rsp: (
+                ZDOCmd.Node_Desc_rsp,
+                {
+                    "Status": rsp.Status,
+                    "NWKAddrOfInterest": rsp.NWK,
+                    "NodeDescriptor": rsp.NodeDescriptor,
+                },
+            )
+        ),
     ),
     ZDOCmd.Active_EP_req: (
         (
@@ -50,19 +59,36 @@ ZDO_CONVERTERS = {
             )
         ),
         (lambda addr: c.ZDO.ActiveEpRsp.Callback(partial=True, Src=addr)),
-        (lambda rsp: (ZDOCmd.Active_EP_rsp, [rsp.ActiveEndpoints])),
+        (
+            lambda rsp: (
+                ZDOCmd.Active_EP_rsp,
+                {
+                    "Status": rsp.Status,
+                    "NWKAddrOfInterest": rsp.NWK,
+                    "ActiveEPList": rsp.ActiveEndpoints,
+                },
+            )
+        ),
     ),
     ZDOCmd.Simple_Desc_req: (
         (
-            # fmt: off
-            lambda addr, device, NWKAddrOfInterest, EndPoint: \
-            c.ZDO.SimpleDescReq.Req(
-                DstAddr=addr, NWKAddrOfInterest=NWKAddrOfInterest, Endpoint=EndPoint
+            lambda addr, device, NWKAddrOfInterest, EndPoint: (
+                c.ZDO.SimpleDescReq.Req(
+                    DstAddr=addr, NWKAddrOfInterest=NWKAddrOfInterest, Endpoint=EndPoint
+                )
             )
-            # fmt: on
         ),
         (lambda addr: c.ZDO.SimpleDescRsp.Callback(partial=True, Src=addr)),
-        (lambda rsp: (ZDOCmd.Simple_Desc_rsp, [rsp.SimpleDescriptor])),
+        (
+            lambda rsp: (
+                ZDOCmd.Simple_Desc_rsp,
+                {
+                    "Status": rsp.Status,
+                    "NWKAddrOfInterest": rsp.NWK,
+                    "SimpleDescriptor": rsp.SimpleDescriptor,
+                },
+            )
+        ),
     ),
     ZDOCmd.Mgmt_Leave_req: (
         (
@@ -73,7 +99,22 @@ ZDO_CONVERTERS = {
             )
         ),
         (lambda addr: c.ZDO.MgmtLeaveRsp.Callback(partial=True, Src=addr)),
-        (lambda rsp: (ZDOCmd.Mgmt_Leave_rsp, [rsp.Status])),
+        (lambda rsp: (ZDOCmd.Mgmt_Leave_rsp, {"Status": rsp.Status})),
+    ),
+    ZDOCmd.Bind_req: (
+        (
+            lambda addr, device, SrcAddress, SrcEndpoint, ClusterID, DstAddress: (
+                c.ZDO.BindReq.Req(
+                    Dst=addr,
+                    Src=SrcAddress,
+                    SrcEndpoint=SrcEndpoint,
+                    ClusterId=ClusterID,
+                    Address=DstAddress,
+                )
+            )
+        ),
+        (lambda addr: c.ZDO.BindRsp.Callback(partial=True, Src=addr)),
+        (lambda rsp: (ZDOCmd.Bind_rsp, {"Status": rsp.Status})),
     ),
 }
 
@@ -107,6 +148,33 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         finally:
             znp.close()
 
+    def _receive_zdo_message(
+        self, cluster: ZDOCmd, *, tsn: t.uint8_t, sender, **zdo_kwargs
+    ) -> None:
+        field_names, field_types = ZDO_CLUSTERS[cluster]
+        assert set(zdo_kwargs) == set(field_names)
+
+        zdo_args = []
+
+        for f_name, f_type in zip(field_names, field_types):
+            # XXX: bugfix for zigpy optional struct bug
+            if f_type.__name__ == "Optional":
+                f_type = f_type.__mro__[1]
+
+            zdo_args.append(f_type(zdo_kwargs[f_name]))
+
+        message = t.serialize_list([t.uint8_t(tsn)] + zdo_args)
+
+        LOGGER.trace("Pretending we received a ZDO message: %s", message)
+        self.handle_message(
+            sender=sender,
+            profile=zigpy.profiles.zha.PROFILE_ID,
+            cluster=cluster,
+            src_ep=ZDO_ENDPOINT,
+            dst_ep=ZDO_ENDPOINT,
+            message=message,
+        )
+
     def on_zdo_relays_message(self, msg: c.ZDO.SrcRtgInd.Callback) -> None:
         LOGGER.info("ZDO device relays: %s", msg)
         device = self.get_device(nwk=msg.DstAddr)
@@ -114,10 +182,18 @@ class ControllerApplication(zigpy.application.ControllerApplication):
 
     def on_zdo_device_announce(self, msg: c.ZDO.EndDeviceAnnceInd.Callback) -> None:
         LOGGER.info("ZDO device announce: %s", msg)
-        self.handle_join(nwk=msg.NWK, ieee=msg.IEEE, parent_nwk=0x0000)
 
-    def on_zdo_device_join(self, msg: c.ZDO.TCDevInd.Callback) -> None:
-        LOGGER.info("ZDO device join: %s", msg)
+        self._receive_zdo_message(
+            cluster=ZDOCmd.Device_annce,
+            tsn=0xFF,
+            sender=self.get_device(ieee=msg.IEEE),
+            NWKAddr=msg.NWK,
+            IEEEAddr=msg.IEEE,
+            Capability=msg.Capabilities,
+        )
+
+    def on_tc_device_join(self, msg: c.ZDO.TCDevInd.Callback) -> None:
+        LOGGER.info("TC device join: %s", msg)
         self.handle_join(nwk=msg.SrcNwk, ieee=msg.SrcIEEE, parent_nwk=msg.ParentNwk)
 
     def on_zdo_device_leave(self, msg: c.ZDO.LeaveInd.Callback) -> None:
@@ -161,7 +237,7 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         )
 
         api.callback_for_response(
-            c.ZDO.TCDevInd.Callback.Callback(partial=True), self.on_zdo_device_join,
+            c.ZDO.TCDevInd.Callback.Callback(partial=True), self.on_tc_device_join,
         )
 
         api.callback_for_response(
@@ -226,7 +302,7 @@ class ControllerApplication(zigpy.application.ControllerApplication):
                 InputClusters=input_clusters,
                 OutputClusters=output_clusters,
             ),
-            RspStatus=t.Status.Success,
+            RspStatus=t.Status.SUCCESS,
         )
 
     async def startup(self, auto_form=False):
@@ -253,12 +329,16 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         await self._znp.nvram_write(NwkNvIds.SRC_RTG_EXPIRY_TIME, t.uint8_t(255))
         await self._znp.nvram_write(NwkNvIds.NWK_CHILD_AGE_ENABLE, t.Bool(False))
 
+        # Reset to make the above NVRAM writes take effect.
+        # This also ensures any previously-started network joins don't continue.
+        await self._reset()
+
         try:
             is_configured = (
                 await self._znp.nvram_read(NwkNvIds.HAS_CONFIGURED_ZSTACK3)
             ) == b"\x55"
         except InvalidCommandResponse as e:
-            assert e.response.Status == t.Status.InvalidParameter
+            assert e.response.Status == t.Status.INVALID_PARAMETER
             is_configured = False
 
         if not is_configured and not auto_form:
@@ -272,13 +352,15 @@ class ControllerApplication(zigpy.application.ControllerApplication):
             dbm = self.config[conf.CONF_ZNP_CONFIG][conf.CONF_TX_POWER]
 
             await self._znp.request(
-                c.Sys.SetTxPower.Req(TXPower=dbm), RspStatus=t.Status.Success
+                c.Sys.SetTxPower.Req(TXPower=dbm), RspStatus=t.Status.SUCCESS
             )
 
         device_info = await self._znp.request(
-            c.Util.GetDeviceInfo.Req(), RspStatus=t.Status.Success
+            c.Util.GetDeviceInfo.Req(), RspStatus=t.Status.SUCCESS
         )
 
+        self._ieee = device_info.IEEE
+
         if device_info.DeviceState != t.DeviceState.StartedAsCoordinator:
             # Start the application and wait until it's ready
             await self._znp.request_callback_rsp(
@@ -292,14 +374,14 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         # Get our active endpoints
         endpoints = await self._znp.request_callback_rsp(
             request=c.ZDO.ActiveEpReq.Req(DstAddr=0x0000, NWKAddrOfInterest=0x0000),
-            RspStatus=t.Status.Success,
+            RspStatus=t.Status.SUCCESS,
             callback=c.ZDO.ActiveEpRsp.Callback(partial=True),
         )
 
         # Clear out the list of active endpoints
         for endpoint in endpoints.ActiveEndpoints:
             await self._znp.request(
-                c.AF.Delete.Req(Endpoint=endpoint), RspStatus=t.Status.Success
+                c.AF.Delete.Req(Endpoint=endpoint), RspStatus=t.Status.SUCCESS
             )
 
         # Register our endpoints
@@ -346,24 +428,24 @@ class ControllerApplication(zigpy.application.ControllerApplication):
 
         if channels is not None:
             await self._znp.request(
-                c.Util.SetChannels.Req(Channels=channels), RspStatus=t.Status.Success,
+                c.Util.SetChannels.Req(Channels=channels), RspStatus=t.Status.SUCCESS,
             )
             await self._znp.request(
                 c.AppConfig.BDBSetChannel.Req(IsPrimary=True, Channel=channels),
-                RspStatus=t.Status.Success,
+                RspStatus=t.Status.SUCCESS,
             )
             await self._znp.request(
                 c.AppConfig.BDBSetChannel.Req(
                     IsPrimary=False, Channel=t.Channels.NO_CHANNELS
                 ),
-                RspStatus=t.Status.Success,
+                RspStatus=t.Status.SUCCESS,
             )
 
             self._channels = channels
 
         if pan_id is not None:
             await self._znp.request(
-                c.Util.SetPanId.Req(PanId=pan_id), RspStatus=t.Status.Success
+                c.Util.SetPanId.Req(PanId=pan_id), RspStatus=t.Status.SUCCESS
             )
 
             self._pan_id = pan_id
@@ -377,7 +459,7 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         if network_key is not None:
             await self._znp.request(
                 c.Util.SetPreConfigKey.Req(PreConfigKey=network_key),
-                RspStatus=t.Status.Success,
+                RspStatus=t.Status.SUCCESS,
             )
 
             # XXX: The Util request does not actually write to this NV address
@@ -427,7 +509,7 @@ class ControllerApplication(zigpy.application.ControllerApplication):
             c.AppConfig.BDBStartCommissioning.Req(
                 Mode=c.app_config.BDBCommissioningMode.NwkFormation
             ),
-            RspStatus=t.Status.Success,
+            RspStatus=t.Status.SUCCESS,
         )
 
         # This may take a while because of some sort of background scanning.
@@ -440,7 +522,7 @@ class ControllerApplication(zigpy.application.ControllerApplication):
             c.AppConfig.BDBStartCommissioning.Req(
                 Mode=c.app_config.BDBCommissioningMode.NwkSteering
             ),
-            RspStatus=t.Status.Success,
+            RspStatus=t.Status.SUCCESS,
         )
 
         # Create the NV item that keeps track of whether or not we're configured
@@ -450,7 +532,7 @@ class ControllerApplication(zigpy.application.ControllerApplication):
             )
         )
 
-        if osal_create_rsp.Status not in (t.Status.Success, t.Status.ItemCreated):
+        if osal_create_rsp.Status not in (t.Status.SUCCESS, t.Status.NV_ITEM_UNINIT):
             raise RuntimeError(
                 "Could not create HAS_CONFIGURED_ZSTACK3 NV item"
             )  # pragma: no cover
@@ -493,38 +575,29 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         device = self.get_device(nwk=dst_addr.address)
 
         # Call the converter with the ZDO request's kwargs
-        req_factory, rsp_factory, zdo_rsp_factory = ZDO_CONVERTERS[cluster]
+        req_factory, rsp_factory, zdo_rsp_factory = ZDO_COMPLEX_CONVERTERS[cluster]
         request = req_factory(dst_addr.address, device, **zdo_kwargs)
         callback = rsp_factory(dst_addr.address)
 
         LOGGER.debug(
-            "Intercepted AP ZDO request and replaced with %s/%s", request, callback
+            "Intercepted AP ZDO request %s(%s) and replaced with %s",
+            cluster,
+            zdo_kwargs,
+            request,
         )
 
         try:
             async with async_timeout.timeout(ZDO_REQUEST_TIMEOUT):
                 response = await self._znp.request_callback_rsp(
-                    request=request, RspStatus=t.Status.Success, callback=callback
+                    request=request, RspStatus=t.Status.SUCCESS, callback=callback
                 )
         except InvalidCommandResponse as e:
             raise DeliveryError(f"Could not send command: {e.response.Status}") from e
 
-        zdo_rsp_cluster, zdo_response_args = zdo_rsp_factory(response)
+        zdo_rsp_cluster, zdo_response_kwargs = zdo_rsp_factory(response)
 
-        # Build up a ZDO response
-        message = t.serialize_list(
-            [t.uint8_t(sequence), response.Status, response.NWK] + zdo_response_args
-        )
-        LOGGER.trace("Pretending we received a ZDO message: %s", message)
-
-        # We do not get any LQI info here
-        self.handle_message(
-            sender=device,
-            profile=zigpy.profiles.zha.PROFILE_ID,
-            cluster=zdo_rsp_cluster,
-            src_ep=dst_ep,
-            dst_ep=src_ep,
-            message=message,
+        self._receive_zdo_message(
+            cluster=zdo_rsp_cluster, tsn=sequence, sender=device, **zdo_response_kwargs
         )
 
         return response.Status, "Request sent successfully"
@@ -555,13 +628,13 @@ class ControllerApplication(zigpy.application.ControllerApplication):
         if dst_addr.mode == t.AddrMode.Broadcast:
             # We won't always get a data confirmation
             response = await self._znp.request(
-                request=request, RspStatus=t.Status.Success
+                request=request, RspStatus=t.Status.SUCCESS
             )
         else:
             async with async_timeout.timeout(DATA_CONFIRM_TIMEOUT):
                 response = await self._znp.request_callback_rsp(
                     request=request,
-                    RspStatus=t.Status.Success,
+                    RspStatus=t.Status.SUCCESS,
                     callback=c.AF.DataConfirm.Callback(
                         partial=True, Endpoint=dst_ep, TSN=sequence
                     ),
@@ -569,7 +642,7 @@ class ControllerApplication(zigpy.application.ControllerApplication):
 
             LOGGER.debug("Received a data request confirmation: %s", response)
 
-        if response.Status != t.Status.Success:
+        if response.Status != t.Status.SUCCESS:
             return response.Status, "Invalid response status"
 
         return response.Status, "Request sent successfully"
@@ -651,33 +724,34 @@ class ControllerApplication(zigpy.application.ControllerApplication):
 
     async def force_remove(self, device) -> None:
         """Forcibly remove device from NCP."""
-        await self._znp.request(
-            c.ZDO.MgmtLeaveReq.Req(
+        leave_rsp = await self._znp.request_callback_rsp(
+            request=c.ZDO.MgmtLeaveReq.Req(
                 DstAddr=0x0000,  # We handle it
                 IEEE=device.ieee,
                 RemoveChildren_Rejoin=c.zdo.LeaveOptions.NONE,
             ),
-            RspStatus=t.Status.Success,
+            RspStatus=t.Status.SUCCESS,
+            callback=c.ZDO.MgmtLeaveRsp.Callback(Src=0x0000, partial=True),
         )
 
+        assert leave_rsp.Status == t.ZDOStatus.SUCCESS
+
         # TODO: see what happens when we forcibly remove a device that isn't our child
 
-        # Just wait for the response, removing the device will be handled upstream
-        await self._znp.wait_for_response(
-            c.ZDO.LeaveInd.Callback(NWK=device.nwk, IEEE=device.ieee, partial=True)
-        )
-
     async def permit_ncp(self, time_s: int) -> None:
         response = await self._znp.request_callback_rsp(
             request=c.ZDO.MgmtPermitJoinReq.Req(
                 AddrMode=t.AddrMode.Broadcast,
                 Dst=zigpy.types.BroadcastAddress.ALL_DEVICES,
                 Duration=time_s,
-                TCSignificance=0,  # not used in Z-Stack
+                # "This field shall always have a value of 1,
+                #  indicating a request to change the
+                #  Trust Center policy."
+                TCSignificance=1,
             ),
-            RspStatus=t.Status.Success,
+            RspStatus=t.Status.SUCCESS,
             callback=c.ZDO.MgmtPermitJoinRsp.Callback(partial=True),
         )
 
-        if response.Status != t.Status.Success:
+        if response.Status != t.Status.SUCCESS:
             raise RuntimeError(f"Permit join response failure: {response}")
